//
//  StreamingPlayer.m
//  iSub
//
//  Created by Aaron Brethorst on 9/23/10.
//  Copyright 2010 Canonical Ltd.
//	
//  This program is free software: you can redistribute it and/or modify it
//  under the terms of the GNU Affero General Public License version 3,
//  as published by the Free Software Foundation.
//	
//  This program is distributed in the hope that it will be useful, but
//  WITHOUT ANY WARRANTY; without even the implied warranties of
//  MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
//  PURPOSE.  See the GNU Affero General Public License for more details.
//
//  You should have received a copy of the GNU Affero General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.

#import "StreamingPlayer.h"
#import "SynthesizeSingleton.h"
#import "Subsonic.h"
#import "Song.h"
#import "NSMutableArray+Extras.h"

#define DEBUG_STREAMER_RETAIN 0

@interface StreamingPlayer ()
- (NSArray*)activePlaylist;
@end

@implementation StreamingPlayer
@synthesize streamer, streamerProgress, repeatMode, isShuffle, isPlaying, seekTime;
@synthesize currentSong, currentPlaylist, shufflePlaylist, currentCoverArt, isNewSong, songUrl, coverArtUrl;

@dynamic currentSongIndex;

SYNTHESIZE_SINGLETON_FOR_CLASS(StreamingPlayer);

- (id)init
{
	if ((self = [super init]))
	{
		repeatMode = RepeatModeOff;
		self.isShuffle = NO;
		isNewSong = NO;
		
		[self addObserver:self forKeyPath:@"isShuffle" options:0 context:NULL];
	}
	return self;
}

- (void)dealloc
{
	[self destroyStreamer];
	[progressUpdateTimer invalidate];
	
	self.currentSong = NULL;
	
	self.currentPlaylist = NULL;
	self.shufflePlaylist = NULL;
	self.currentCoverArt = NULL;
	
	self.songUrl = NULL;
	self.coverArtUrl = NULL;
	
	[super dealloc];
}

- (void)recoverState:(BOOL)withShuffle
{
	//AB - I'm not a big fan of pulling state from the user defaults in this class.
	//Rather do this in a cleaner way, but this'll have to suffice for now.
	NSUserDefaults *defaults = [NSUserDefaults standardUserDefaults];
	
	if ([[defaults objectForKey:@"isShuffle"] isEqualToString:@"YES"])
	{
		self.isShuffle = YES;
		[self setShufflePlaylist:[defaults objectForKey:@"shufflePlaylist"]];
		[self setCurrentPlaylist:[defaults objectForKey:@"currentPlaylist"]];
	}
	else 
	{
		self.isShuffle = NO;
		[self setCurrentPlaylist:[defaults objectForKey:@"currentPlaylist"]];
	}
	
	self.repeatMode = [[defaults objectForKey:@"repeatMode"] integerValue];
	self.currentSong = [NSKeyedUnarchiver unarchiveObjectWithData:[defaults objectForKey:@"currentSong"]];
	self.currentCoverArt = [UIImage imageWithData:[defaults objectForKey:@"currentCoverArt"]];
	
	self.seekTime = [[defaults objectForKey:@"seekTime"] floatValue];
	self.isPlaying = YES;
	
	self.songUrl = [[Subsonic sharedSubsonic] getStreamingURLForSongId:self.currentSong.songId];
	[self createStreamer];
	[self.streamer seekToTime:self.seekTime];
}

- (void)saveDefaults
{
	if (self.isPlaying)
	{
		NSUserDefaults *defaults = [NSUserDefaults standardUserDefaults];
		if (self.isShuffle)
		{
			[defaults setObject:@"YES" forKey:@"isShuffle"];
			[defaults setObject:shufflePlaylist forKey:@"shufflePlaylist"];
			[defaults setObject:currentPlaylist forKey:@"currentPlaylist"];
		}
		else
		{
			[defaults setObject:@"NO" forKey:@"isShuffle"];
			[defaults setObject:currentPlaylist forKey:@"currentPlaylist"];
		}
		
		[defaults setObject:[NSString stringWithFormat:@"%i", repeatMode] forKey:@"repeatMode"];
		[defaults setObject:[NSKeyedArchiver archivedDataWithRootObject:currentSong] forKey:@"currentSong"];
		
		[defaults setObject:UIImageJPEGRepresentation(currentCoverArt, 1.0) forKey:@"currentCoverArt"];
		[self.streamer progress];
		[defaults setObject:[NSString stringWithFormat:@"%f", (seekTime + streamerProgress)] forKey:@"seekTime"];
		[defaults setObject:@"YES" forKey:@"recover"];
		[defaults synchronize];
	}
}

#pragma mark -
#pragma mark KVO

- (void)observeValueForKeyPath:(NSString *)keyPath ofObject:(id)object change:(NSDictionary *)change context:(void *)context
{
	if ([keyPath isEqual:@"isShuffle"] && self.isShuffle)
	{
		NSMutableArray *shuffle = [NSMutableArray arrayWithArray:self.currentPlaylist];
		[shuffle removeObject:self.currentSong];
		
		[shuffle shuffle];
		[shuffle insertObject:self.currentSong atIndex:0];
		self.shufflePlaylist = shuffle;
	}
}

#pragma mark -
#pragma mark Music Streamer

- (void)createStreamer
{
#if DEBUG_STREAMER_RETAIN
	NSLog(@"- (void)createStreamer");
#endif
	
	@synchronized(self)
	{
		[self destroyStreamer];
		
		AudioStreamer *as = nil;
		
		if ([self.currentSong cachedSongExists])
		{
			as = [[AudioStreamer alloc] initWithFileURL:[self.currentSong cachedSongURL]];
		}
		else
		{
			as = [[AudioStreamer alloc] initWithURL:songUrl];
		}

		self.streamer = as;
		[as release];
#if DEBUG_STREAMER_RETAIN
		NSLog(@"New streamer is %x", self.streamer);
#endif
		[[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(playbackStateChanged:) name:ASStatusChangedNotification object:self.streamer];
	}
}


- (void)destroyStreamer
{
#if DEBUG_STREAMER_RETAIN
	NSLog(@"- (void)destroyStreamer");
#endif
	@synchronized(self)
	{
		if (self.streamer)
		{
			[[NSNotificationCenter defaultCenter] removeObserver:self name:ASStatusChangedNotification object:self.streamer];
#if DEBUG_STREAMER_RETAIN
			NSLog(@"Destroying %x (RETAIN %d)", self.streamer, [self.streamer retainCount]);
#endif
			self.streamer = NULL;
		}
	}
}


- (void)playPauseSong
{
	if ([self.streamer isPaused])
	{
		self.isPlaying = YES;
		[self.streamer start];
	}
	else if ([self.streamer isPlaying])
	{
		self.isPlaying = NO;
		[self.streamer pause];
	}
	else 
	{
		self.isPlaying = YES;
		seekTime = 0.0;
		[self createStreamer];
		[self.streamer start];
	}
}

- (int)currentSongIndex
{
	return [[self activePlaylist] indexOfObject:self.currentSong];
}

- (NSArray*)activePlaylist
{
	return (self.isShuffle ? self.shufflePlaylist : self.currentPlaylist);
}

- (void)prevSong
{
	[self jumpAhead:-1];
}

- (void)jumpAhead:(int)by
{
	@synchronized(self)
	{
		NSArray *playlist = [self activePlaylist];
		int index = [playlist indexOfObject:currentSong];
		
		index = index + by; //index of the next song.
		
		if (index < [playlist count])
		{
#if DEBUG_STREAMER_RETAIN
			NSLog(@"Moving from %@ (%d) to %@ (%d)", self.currentSong.title, index-1, [[playlist objectAtIndex:index] title], index);
#endif
			self.currentSong = [playlist objectAtIndex:index];
			
			self.songUrl = [[Subsonic sharedSubsonic] getStreamingURLForSongId:self.currentSong.songId];
			
			[self.streamer stop];
			[self playPauseSong];
		}	
		else
		{
			self.isPlaying = NO;
			[self destroyStreamer];
		}
	}	
}

- (void)nextSong
{
	[self jumpAhead:1];
}


- (void)nextSongAuto
{
	// If it's in regular play mode, then go to the next track.
	if (RepeatModeOff == repeatMode)
	{
		[self nextSong];
	}
	// If it's in repeat-one mode then just restart the streamer
	else if (RepeatModeOne == repeatMode)
	{
		[self.streamer stop];
		[self playPauseSong];
	}
	// If it's in repeat-all mode then check if it's at the end of the playlist and start from the beginning, or just go to the next track.
	else if (RepeatModeAll == repeatMode)
	{
		NSInteger index = [self.currentPlaylist indexOfObject:self.currentSong];
		index = index + 1;
		
		if (index < [self.currentPlaylist count])
		{
			[self nextSong];
		}
		else
		{
			NSArray *playlist;
			
			playlist = self.isShuffle ? [NSArray arrayWithArray:shufflePlaylist] : [NSArray arrayWithArray:currentPlaylist];
			
			self.currentSong = [playlist objectAtIndex:0];
			
			self.songUrl = [[Subsonic sharedSubsonic] getStreamingURLForSongId:self.currentSong.songId];
			
			[self.streamer stop];
			[self playPauseSong];
		}
	}
}

- (void)playbackStateChanged:(NSNotification *)aNotification
{
	if ([self.streamer isWaiting])
	{
		//maybe a nice buffering animation could be played in the view code...
	}
	else if ([self.streamer isPlaying])
	{
		self.isPlaying = YES;
	}
	else if ([self.streamer isIdle])
	{
		[self nextSongAuto];
	}
}

@end