//
//  AlbumArtLoader.m
//  iSub
//
//  Created by Aaron Brethorst on 5/4/11.
//  Copyright 2011 Canonical Ltd.
//	
//  This program is free software: you can redistribute it and/or modify it
//  under the terms of the GNU Affero General Public License version 3,
//  as published by the Free Software Foundation.
//	
//  This program is distributed in the hope that it will be useful, but
//  WITHOUT ANY WARRANTY; without even the implied warranties of
//  MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
//  PURPOSE.  See the GNU Affero General Public License for more details.
//
//  You should have received a copy of the GNU Affero General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.

#import "AlbumArtLoader.h"
#import "Subsonic.h"
#import "NSMutableArray+Extras.h"
#import "UIImage+Resize.h"

@interface AlbumArtLoader ()
- (void)createAlbumArtDirectory;
- (NSString*)URLFromCoverArtID:(NSString*)artID;
@end


@implementation AlbumArtLoader
@synthesize delegate;
@synthesize imageSize;
@synthesize defaultAlbumArt;

- (id)init
{
	if ((self = [super init]))
	{
        [self createAlbumArtDirectory];
        
		self.imageSize = 640; // Default to maximum size we need
        self.defaultAlbumArt = [UIImage imageNamed:@"default-album-art"];
		
		downloadCache = [[NSMutableDictionary alloc] init];
		downloadedURLs = [[NSMutableSet alloc] init];
		queue = [[NSOperationQueue alloc] init];
		[queue setSuspended:NO];
	}
	return self;
}

- (void)dealloc
{
	[downloadCache release];
	[downloadedURLs release];
	[queue release];
    [defaultAlbumArt release];
	
	[super dealloc];
}

#pragma mark -
#pragma mark Public Methods

- (UIImage*)imageForArtID:(NSString*)artID error:(NSError**)error
{
    NSString *path = CachedAlbumArtPathForArtId(artID);
	UIImage *image = [UIImage imageWithContentsOfFile:path];

    if (nil == image) {
        // Cache fault: return the default image, and set the error so the caller can dispatch for an attempt at download
        image = self.defaultAlbumArt;
        *error = [NSError errorWithDomain:@"U1Music" code:1 userInfo:nil];
    }
	return [image resizedImage:CGSizeMake(self.imageSize, self.imageSize) interpolationQuality:kCGInterpolationHigh];
}

- (void)download:(NSString*)artID
{	
	if (nil == artID)
	{
		return;
	}
	
	NSURL *url = [NSURL URLWithString:[self URLFromCoverArtID:artID]];
	
	@synchronized(self)
	{
		if (![downloadedURLs containsObject:url])
		{
			[downloadedURLs addObject:url];
			
			AlbumArtLoadingOperation *operation = [[AlbumArtLoadingOperation alloc] initWithArtId: artID URL:url];
			operation.delegate = self;
			[queue addOperation:operation];
			[operation release];
		}
	}
}

+ (NSString*)cacheSize
{
	static int minBlock = 4096;
	
	NSString *file = nil;
	NSDictionary *fileAttrs = nil;
	
	double totalSize = 0;
	
	NSDirectoryEnumerator *enumerator = [[NSFileManager defaultManager] enumeratorAtPath:[CachedAlbumArtDirectory() path]];
	
	while (file = [enumerator nextObject])
	{
		fileAttrs = [enumerator fileAttributes];
		
		if (![[fileAttrs valueForKey:NSFileType] isEqual:NSFileTypeDirectory])
		{
			totalSize += ((([[fileAttrs valueForKey:NSFileSize] unsignedIntValue] + minBlock - 1) / minBlock) * minBlock);
		}
	}
	
	return [NSString stringWithFormat:@"%.2f MB",((totalSize / 1024.0f) / 1024.0f)];
}

#pragma mark -
#pragma mark Private Methods

- (NSString*)URLFromCoverArtID:(NSString*)artID
{
	NSMutableArray *parameters = [NSMutableArray array];
	[parameters addKeyValueObjectFromArray:[NSArray arrayWithObjects:@"id", artID, nil]];

	// Always get the largest we can handle
	[parameters addKeyValueObjectFromArray:[NSArray arrayWithObjects:@"size", [NSString stringWithFormat:@"%d", 640], nil]];
	
	NSURL *url = [[Subsonic sharedSubsonic] getBaseURL:@"getCoverArt.view" parameters:parameters];
	return [url absoluteString];
}

- (void)createAlbumArtDirectory
{
	NSError *error = nil;
	BOOL yn = [[NSFileManager defaultManager] createDirectoryAtPath:[CachedAlbumArtDirectory() path] withIntermediateDirectories:YES attributes:nil error:&error];
	
	if (!yn || error)
	{
		NSLog(@"%@", error);
	}
}

#pragma mark -
#pragma mark AlbumArtLoadingOperationDelegate

- (void)imageLoaded:(UIImage*)image forArtId:(NSString *)artId
{
    NSData *imageData = UIImagePNGRepresentation(image);
    [imageData writeToFile:CachedAlbumArtPathForArtId(artId) atomically:YES];
	
	if (self.delegate)
	{
		[self.delegate performSelectorOnMainThread:@selector(reloadImages) withObject:nil waitUntilDone:NO];
	}
}

@end
