//
//  Downloader.m
//  iSub
//
//  Created by Aaron Brethorst on 5/12/11.
//  Copyright 2011 Canonical Ltd.
//	
//  This program is free software: you can redistribute it and/or modify it
//  under the terms of the GNU Affero General Public License version 3,
//  as published by the Free Software Foundation.
//	
//  This program is distributed in the hope that it will be useful, but
//  WITHOUT ANY WARRANTY; without even the implied warranties of
//  MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
//  PURPOSE.  See the GNU Affero General Public License for more details.
//
//  You should have received a copy of the GNU Affero General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.

#import "Downloader.h"
#import "DownloadOperation.h"
#import "SynthesizeSingleton.h"
#import "Artist.h"
#import "Album.h"
#import "Song.h"
#import "Subsonic.h"

static const NSUInteger kMaxConcurrentDownloads = 4;

@interface Downloader ()
- (void)createMusicDirectory;

- (void)startArtistEnqueueing:(Artist*)anArtist;
- (void)completeArtistEnqueueing:(Artist*)anArtist;

- (void)startAlbumEnqueueing:(Album*)anAlbum;
- (void)completeAlbumEnqueueing:(Album*)anAlbum;
@end

@implementation Downloader

SYNTHESIZE_SINGLETON_FOR_CLASS(Downloader);

- (id)init
{
	if ((self = [super init]))
	{
		[self createMusicDirectory];
		
		downloads = [[NSMutableSet alloc] init];
		
		queue = [[NSOperationQueue alloc] init];
		[queue setMaxConcurrentOperationCount:kMaxConcurrentDownloads];
		[queue setSuspended:NO];
		
		contentEnumerationQueue = [[NSOperationQueue alloc] init];
		[contentEnumerationQueue setSuspended:NO];
	}
	return self;
}

- (void)dealloc
{
    [contentEnumerationQueue release];
	[downloads release];
	[queue release];
	
	[super dealloc];
}

#pragma mark -
#pragma mark Public Methods

- (void)enqueueArtist:(Artist*)anArtist
{
	NSInvocationOperation *invOp = [[[NSInvocationOperation alloc] initWithTarget:self selector:@selector(startArtistEnqueueing:) object:anArtist] autorelease];
	[contentEnumerationQueue addOperation:invOp];
}

- (void)enqueueAlbum:(Album*)anAlbum
{
	NSInvocationOperation *invOp = [[[NSInvocationOperation alloc] initWithTarget:self selector:@selector(startAlbumEnqueueing:) object:anAlbum] autorelease];
	[contentEnumerationQueue addOperation:invOp];
}

- (void)downloadFile:(NSURL*)url withName:(NSString*)fileName
{
	[self downloadFile:url withName:fileName completionBlock:NULL];
}

- (void)downloadFile:(NSURL*)url withName:(NSString*)fileName completionBlock:(void(^)(NSString *path))completionBlock
{
	@synchronized(self)
	{
		if ([downloads containsObject:fileName])
		{
			return;
		}
		
		NSString *path = CachedMusicPathForFilename(fileName);

		if (CacheContainsFile(path))
		{
			// Partially downloaded file, delete and try again
			[[NSFileManager defaultManager] removeItemAtPath:path error:NULL];
		}
		
		[downloads addObject:fileName];
		// Disable idle timer sleep while downloading
		[[UIApplication sharedApplication] setIdleTimerDisabled:YES];
		
		DownloadOperation *download = [DownloadOperation downloadOperationWithURL:url
																		 filePath:path];
		download.delegate = self;
		[download setCompletionBlock:^(void) {
			if (completionBlock)
				completionBlock(path);
		}];
		[queue addOperation:download];
	}
}

- (BOOL)isDownloading:(NSString *)fileName
{
    return [downloads containsObject:fileName];
}

+ (NSString*)cacheSize
{
	static int minBlock = 4096;
	
	NSString *file = nil;
	NSDictionary *fileAttrs = nil;
	
	double totalSize = 0;
	
	NSDirectoryEnumerator *enumerator = [[NSFileManager defaultManager] enumeratorAtPath:[CachedMusicDirectory() path]];
	
	while (file = [enumerator nextObject])
	{
		fileAttrs = [enumerator fileAttributes];
		
		if (![[fileAttrs valueForKey:NSFileType] isEqual:NSFileTypeDirectory])
		{
			totalSize += ((([[fileAttrs valueForKey:NSFileSize] unsignedIntValue] + minBlock - 1) / minBlock) * minBlock);
		}
	}
	
	return [NSString stringWithFormat:@"%.2f MB",((totalSize / 1024.0f) / 1024.0f)];
}

#pragma mark -
#pragma mark DownloadOperationDelegate

- (void)songDownloadedToPath:(NSString*)path
{
	[downloads removeObject:[path lastPathComponent]];
    if ([downloads count] == 0)
    {
        // If downloads is empty, go back to idle timing for sleep
        [[UIApplication sharedApplication] setIdleTimerDisabled:NO];
    }
    // Coarse-grained, blanket notification that any download has completed.
    [[NSNotificationCenter defaultCenter] postNotificationName:NOTIF_downloadComplete object:nil];
}

#pragma mark -
#pragma mark Private Methods

- (void)startArtistEnqueueing:(Artist*)anArtist
{
	[anArtist loadAlbums];
	[self performSelectorOnMainThread:@selector(completeArtistEnqueueing:) withObject:anArtist waitUntilDone:NO];
}

- (void)completeArtistEnqueueing:(Artist*)anArtist
{
	[PerThreadManagedObjectContext() refreshObject:anArtist mergeChanges:YES]; //force the artist to reload its set of albums from the persistent store.
	
	for (Album *a in anArtist.albums)
	{
		[self enqueueAlbum:a];
	}
}

- (void)startAlbumEnqueueing:(Album*)anAlbum
{
	[anAlbum loadSongs];
	[self performSelectorOnMainThread:@selector(completeAlbumEnqueueing:) withObject:anAlbum waitUntilDone:NO];
}

- (void)completeAlbumEnqueueing:(Album*)anAlbum
{
	[PerThreadManagedObjectContext() refreshObject:anAlbum mergeChanges:YES]; //force the album to reload its set of songs from the persistent store.

	for (Song *s in anAlbum.songs)
	{
		[self downloadFile:[[Subsonic sharedSubsonic] getStreamingURLForSongId:s.songId] withName:s.songId];
	}
}

- (void)createMusicDirectory
{
	NSError *error = nil;
	BOOL yn = [[NSFileManager defaultManager] createDirectoryAtPath:[CachedMusicDirectory() path] withIntermediateDirectories:YES attributes:nil error:&error];
	
	if (!yn || error)
	{
		NSLog(@"%@", error);
	}
}

@end
