//
//  AlbumListViewController.m
//  iSub
//
//  Created by Zachery Bir on 3/16/11.
//  Copyright 2011 Canonical Ltd.
//	
//  This program is free software: you can redistribute it and/or modify it
//  under the terms of the GNU Affero General Public License version 3,
//  as published by the Free Software Foundation.
//	
//  This program is distributed in the hope that it will be useful, but
//  WITHOUT ANY WARRANTY; without even the implied warranties of
//  MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
//  PURPOSE.  See the GNU Affero General Public License for more details.
//
//  You should have received a copy of the GNU Affero General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.

#import "AlbumListViewController.h"
#import "AlbumListParser.h"
#import "AlbumArtistUITableViewCell.h"
#import "Album.h"
#import "AlbumViewController.h"
#import "NSString+Extras.h"

@interface AlbumListViewController ()
- (void)configureCell:(AlbumArtistUITableViewCell*)cell forTableView:(UITableView*)tableView indexPath:(NSIndexPath*)indexPath;
@end

@implementation AlbumListViewController

+ (UINavigationController *)navigableViewController
{
    AlbumListViewController *albumListViewController = [[[AlbumListViewController alloc] initWithTitle:NSLocalizedString(@"Albums", @"")] autorelease];
	AlbumListParser *parser = [[AlbumListParser alloc] init];
	albumListViewController.parserDelegate = parser;
	albumListViewController.viewName = @"getAlbums.view";
	
    UINavigationController *nav = [[[UINavigationController alloc] initWithRootViewController:albumListViewController] autorelease];
    nav.navigationBar.barStyle = UIBarStyleBlack;
    nav.tabBarItem.image = [UIImage imageNamed:@"albums"];
    return nav;
}

#pragma mark - View stuff

- (void)viewDidLoad
{
	self.tableView.rowHeight = 60.0f;
    [super viewDidLoad];
	
	loader = [[AlbumArtLoader alloc] init];
	loader.delegate = self;
	loader.imageSize = 120;
}

- (void)viewDidUnload;
{
    [super viewDidUnload];
    loader.delegate = nil;
    [loader release], loader = nil;
}

#pragma mark - Data Loading

- (void)loadLocalData
{
	@synchronized(self.tableData)
	{
		[self.tableData removeAllObjects];
		self.groupedTableData = [NSMutableDictionary dictionary];
		self.indexes = [NSArray array];
		
		[self.tableData addObjectsFromArray:[PerThreadManagedObjectContext() fetchObjectsForEntityName:@"Album" withPredicate:nil]];
		
		NSMutableSet *set = [NSMutableSet set];
		
		for (Album *a in self.tableData)
		{
			NSString *index = [[a.title dearticlizedIndex] uppercaseString];
			NSMutableArray *group = [self.groupedTableData objectForKey:index];
			
			if (nil == group)
			{
				group = [NSMutableArray array];
				[self.groupedTableData setObject:group forKey:index];
			}
			
			[group addObject:a];
			[set addObject:index];
		}
		
		self.indexes = [[set allObjects] sortedArrayUsingSelector:@selector(compare:)];
		
		indexed = ([self.tableData count] > 8);
	}
	[self.tableView reloadData];
}


- (void)updateReachability:(Reachability*)reachability;
{
	[super updateReachability:reachability];
	
	for (AlbumArtistUITableViewCell *cell in [self.tableView visibleCells])
	{
		[self configureCell:cell forTableView:self.tableView indexPath:[self.tableView indexPathForCell:cell]];
	}
}

#pragma mark - UITableView

- (UITableViewCell *)tableView:(UITableView *)tableView cellForRowAtIndexPath:(NSIndexPath *)indexPath 
{	
	static NSString *AlbumArtistCellIdentifier = @"AlbumArtistCellIdentifier";
	AlbumArtistUITableViewCell *cell = (AlbumArtistUITableViewCell *)[tableView dequeueReusableCellWithIdentifier:AlbumArtistCellIdentifier];
	
	if (nil == cell)
	{
		cell = [[[AlbumArtistUITableViewCell alloc] initWithStyle:UITableViewCellStyleDefault reuseIdentifier:AlbumArtistCellIdentifier] autorelease];
	}
	[self configureCell:cell forTableView:tableView indexPath:indexPath];
	return cell;
}

- (void)configureCell:(AlbumArtistUITableViewCell*)cell forTableView:(UITableView*)tableView indexPath:(NSIndexPath*)indexPath;
{
    Album *album = nil;
    
	if (self.searchDisplayController.searchResultsTableView == tableView)
	{
		album = [self.searchResults objectAtIndex:indexPath.row];
		cell.accessoryType = UITableViewCellAccessoryDisclosureIndicator;
	}
	else
	{
		cell.accessoryType = UITableViewCellAccessoryNone;

        if (indexed)
        {
            NSString *index = [self.indexes objectAtIndex:indexPath.section];
            album = [[self.groupedTableData objectForKey:index] objectAtIndex:indexPath.row]; 
        }
        else
        {
            album = [self.tableData objectAtIndex:indexPath.row];
        }
	}

    cell.albumNameLabel.text = album.title;
    cell.artistNameLabel.text = album.artist;
	
    NSError *error = nil;
    cell.coverArtView.image = [loader imageForArtID:album.coverArtId error:&error];
    
    if ([error code]) {
        // Let's see if we need to wait for non-dragging/decelerating
//        if (self.tableView.dragging == NO && self.tableView.decelerating == NO)
//        {
            [loader download:album.coverArtId];
//        }
    }
	if (canStream || [album hasCachedSongs])
	{
		cell.albumNameLabel.textColor = [UIColor blackColor];
	}
	else
	{
		cell.albumNameLabel.textColor = [UIColor lightGrayColor];
		cell.accessoryType = UITableViewCellAccessoryNone;
	}
}

- (void)tableView:(UITableView *)tableView didSelectRowAtIndexPath:(NSIndexPath *)indexPath
{
	Album *selection = nil;
		
	if (self.searchDisplayController.searchResultsTableView == tableView)
	{
		selection = [self.searchResults objectAtIndex:indexPath.row];
	}
	else if(indexed) 
	{	
		NSString *index = [self.indexes objectAtIndex:indexPath.section];
		selection = [[self.groupedTableData objectForKey:index] objectAtIndex:indexPath.row];
	}
    else 
    {
        selection = [self.tableData objectAtIndex:indexPath.row];
    }

	if (canStream || [selection hasCachedSongs])
	{
		AlbumViewController *albumViewController = [[AlbumViewController alloc] initWithStyle:UITableViewStylePlain];
		albumViewController.album = selection;
		albumViewController.artist = selection.artist;
		[self.navigationController pushViewController:albumViewController animated:YES];
		[albumViewController release];
	}
	else
	{
		[tableView deselectRowAtIndexPath:indexPath animated:YES];
	}
}


#pragma mark -
#pragma mark Deferred Image Loading

- (void)reloadImages
{
	[self.tableView reloadData];
}

- (void)loadImagesForOnscreenRows
{
	for (NSIndexPath *indexPath in [self.tableView indexPathsForVisibleRows])
	{
		Album *album = [self.tableData objectAtIndex:indexPath.row];
        // This will prompt the album art to download, but doesn't try to load the image when it's done?
		[loader download:album.coverArtId];
	}
}

- (void)scrollViewDidEndDragging:(UIScrollView *)scrollView willDecelerate:(BOOL)decelerate
{	
	[super scrollViewDidEndDragging:scrollView willDecelerate:decelerate];
	
    if (!decelerate)
	{
        [self loadImagesForOnscreenRows];
    }
}

- (void)scrollViewDidEndDecelerating:(UIScrollView *)scrollView
{	
    [self loadImagesForOnscreenRows];
}

@end
