//
//  AlbumViewController.m
//  iSub
//
//  Created by Aaron Brethorst on 8/25/10.
//  Copyright 2010 Canonical Ltd.
//	
//  This program is free software: you can redistribute it and/or modify it
//  under the terms of the GNU Affero General Public License version 3,
//  as published by the Free Software Foundation.
//	
//  This program is distributed in the hope that it will be useful, but
//  WITHOUT ANY WARRANTY; without even the implied warranties of
//  MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
//  PURPOSE.  See the GNU Affero General Public License for more details.
//
//  You should have received a copy of the GNU Affero General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.

#import "AlbumViewController.h"

#import "AlbumUITableViewCell.h"
#import "AlbumParser.h"
#import "Album.h"
#import "Song.h"
#import "SongUITableViewCell.h"
#import "NSDate+Extras.h"
#import "StreamingPlayer.h"
#import "Downloader.h"
#import "UIAlertView+Blocks.h"
#import "RIButtonItem.h"

@interface AlbumViewController (Private)
- (void)setupHeaderAndFooter;
- (void)updateNowPlayingImage:(id)sender;
- (void)toggleCacheForAlbum:(id)sender;
- (void)toggleCacheForSong:(id)sender;
- (void)configureCell:(SongUITableViewCell*)cell forSongAtIndexPath:(NSIndexPath*)indexPath;
@end

@implementation AlbumViewController
@synthesize artist, album, songs, cacheButton;

- (id)initWithStyle:(UITableViewStyle)aStyle
{
	if ((self = [super initWithStyle:aStyle]))
	{
		self.songs = [NSMutableArray array];
	}
	return self;
}

#pragma mark -
#pragma mark View lifecycle

- (void)viewDidLoad
{
	[super viewDidLoad];
	
	self.title = self.album.title;
	
	loader = [[AlbumArtLoader alloc] init];
	loader.delegate = self;
	loader.imageSize = 200;

	[self setupHeaderAndFooter];
}

- (void)viewDidAppear:(BOOL)animated
{
	[super viewDidAppear:animated];
	
	[self.tableView reloadData];
}

#pragma mark -
#pragma mark Data Loader

- (void)loadLocalData
{
	@synchronized(self.songs)
	{
		[self.songs removeAllObjects];
		[PerThreadManagedObjectContext() refreshObject:self.album mergeChanges:YES]; //force the album to reload its set of songs from the persistent store.
		[self.songs addObjectsFromArray:[self.album.songs allObjects]];
		[self.songs sortUsingSelector:@selector(compare:)];
	}
	
	[self.tableView reloadData];
}

- (void)beginLoadingRemoteData:(NSString*)force
{
	NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
	
	NSError *error = [self.album loadSongs];
	
	if (error)
	{
		[self performSelectorOnMainThread:@selector(failedLoadingData:) withObject:error waitUntilDone:NO];
	}
		
	[self performSelectorOnMainThread:@selector(loadLocalData) withObject:nil waitUntilDone:NO];
	[self performSelectorOnMainThread:@selector(finishLoadingData) withObject:nil waitUntilDone:NO];
	[pool release];
}

- (void)finishLoadingData
{
	[super finishLoadingData];
	[[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(updateNowPlayingImage:) name:ASStatusChangedNotification object:nil];
}

#pragma mark -
#pragma mark Table view data source

- (NSInteger)numberOfSectionsInTableView:(UITableView *)tableView
{
	return 1;
}


- (NSInteger)tableView:(UITableView *)tableView numberOfRowsInSection:(NSInteger)section
{
	return [self.songs count];
}

- (UITableViewCell *)tableView:(UITableView *)tv cellForRowAtIndexPath:(NSIndexPath *)indexPath
{
	static NSString *SongCellIdentifier = @"SongCellIdentifier";
	SongUITableViewCell *cell = (SongUITableViewCell*)[tv dequeueReusableCellWithIdentifier:SongCellIdentifier];
	
	if (nil == cell)
	{
		cell = [[[SongUITableViewCell alloc] initWithStyle:UITableViewCellStyleDefault reuseIdentifier:SongCellIdentifier] autorelease];
	}
	[self configureCell:cell forSongAtIndexPath:indexPath];
	return cell;
}

- (void)configureCell:(SongUITableViewCell*)cell forSongAtIndexPath:(NSIndexPath*)indexPath
{
	Song *song = [self.songs objectAtIndex:indexPath.row];
	
    // explicitly re-set the default state of the button/spinner
    [cell.cacheButton setImage:(canCache ? [UIImage imageNamed:@"uncached"] : [UIImage imageNamed:@"uncached-disabled"]) forState:UIControlStateNormal];
    [cell.spinner stopAnimating];
    
    cell.cacheButton.tag = indexPath.row;
    [cell.cacheButton addTarget:self action:@selector(toggleCacheForSong:) forControlEvents:UIControlEventTouchUpInside];
    
    if ([[Downloader sharedDownloader] isDownloading:song.songId])
    {
        [cell.spinner startAnimating];
    }
    else if ([song cachedSongExists])
    {
        [cell.cacheButton setImage:[UIImage imageNamed:@"cached"] forState:UIControlStateNormal];
    }
    
	cell.songNameLabel.text = song.title;
	
	cell.positionLabel.text = [NSString stringWithFormat:@"%@", [song.track intValue] > 0 ? [song.track stringValue] : @""];
	
	if (song.duration && [song.duration intValue] > 0)
	{
		cell.durationLabel.text = [NSDate stringFromDuration:song.duration];
	}
	else
	{
		cell.durationLabel.text = @"";
	}
	
	if ([StreamingPlayer sharedStreamingPlayer].currentSong && [song.songId isEqual:[StreamingPlayer sharedStreamingPlayer].currentSong.songId])
	{
		cell.nowPlaying.hidden = NO;
	}
	else
	{
		cell.nowPlaying.hidden = YES;
	}
	
	cell.evenRow = (0 == indexPath.row % 2);
	if (song.cachedSongPath == nil && !canStream)
	{
		cell.songNameLabel.textColor = [UIColor lightGrayColor];
		cell.positionLabel.textColor = [UIColor lightGrayColor];
		cell.durationLabel.textColor = [UIColor lightGrayColor];
	}
	else
	{
		cell.songNameLabel.textColor = [UIColor blackColor];
		cell.positionLabel.textColor = [UIColor blackColor];
		cell.durationLabel.textColor = [UIColor blackColor];
	}
}


#pragma mark -
#pragma mark Table view delegate

- (void)tableView:(UITableView *)tableView didSelectRowAtIndexPath:(NSIndexPath *)indexPath
{
	Song *song = [self.songs objectAtIndex:indexPath.row];
	if ([StreamingPlayer sharedStreamingPlayer].currentSong && [song.songId isEqual:[StreamingPlayer sharedStreamingPlayer].currentSong.songId])
	{
		[self nowPlayingAction:self];
	}
    else
    {
		if (canStream || song.cachedSongPath)
			[self didSelectSong:song withPlaylist:self.songs];
    }
}

#pragma mark -
#pragma mark Memory management

- (void)didReceiveMemoryWarning
{
    [super didReceiveMemoryWarning];
	self.songs = NULL;
}

- (void)viewDidUnload
{
	[super viewDidUnload];
	loader.delegate = nil;
    [loader release], loader = nil;
	[[NSNotificationCenter defaultCenter] removeObserver:self];
	RELEASE_SAFELY(headerView);
}

- (void)dealloc
{
	self.songs = NULL;
	RELEASE_SAFELY(albumLabel);
	RELEASE_SAFELY(artistLabel);
	RELEASE_SAFELY(coverArt);
    RELEASE_SAFELY(cacheButton);
	self.artist = NULL;
	self.album = NULL;
    [super dealloc];
}

- (void)reloadImages
{
	return; // Not sure what to do with the header in this case.
}

- (void)resetTableData:(id)sender
{
    [super resetTableData:sender];
    self.album = nil;
    [self.tableView reloadData];
}

- (void)updateReachability:(Reachability *)reachability;
{
	[super updateReachability:reachability];
	for (SongUITableViewCell *cell in [self.tableView visibleCells])
	{
		[self configureCell:cell forSongAtIndexPath:[self.tableView indexPathForCell:cell]];
	}
}

@end

@implementation AlbumViewController (Private)

- (void)toggleCacheForSong:(id)sender
{
    // This is a binary toggle, and right now, we don't bother trying to cancel a download in progress.
    int row = [(UIButton *)sender tag];

	NSIndexPath *songPath = [NSIndexPath indexPathForRow:row inSection:0];
    SongUITableViewCell *cell = (SongUITableViewCell *)[self.tableView cellForRowAtIndexPath:songPath];
    Song *song = [self.songs objectAtIndex:row];

    if (![[Downloader sharedDownloader] isDownloading:song.songId])
        // If we're downloading at all, don't do anything.
    {
        if ([song cachedSongExists])
        {
            [cell.spinner stopAnimating];
            NSError *error = nil;
            [[NSFileManager defaultManager] removeItemAtPath:CachedMusicPathForFilename(song.songId) error:&error];
            [[NSNotificationCenter defaultCenter] postNotificationName:NOTIF_removeCachedContent object:nil];
            [cell.cacheButton setImage:[UIImage imageNamed:@"uncached"] forState:UIControlStateNormal];
			song.cachedSongPath = nil;
			SaveContext();
        }
        else
        {
			if (canCache)
			{
				[cell.spinner startAnimating];
				[[Downloader sharedDownloader] downloadFile:[[Subsonic sharedSubsonic] getStreamingURLForSongId:song.songId] withName:song.songId completionBlock:^(NSString *path) {
					dispatch_async(dispatch_get_main_queue(), ^(void) {
						song.cachedSongPath = path;
						SaveContext();
						SongUITableViewCell *cell = (SongUITableViewCell *)[self.tableView cellForRowAtIndexPath:songPath];
						[cell.spinner stopAnimating];
						[cell.cacheButton setImage:[UIImage imageNamed:@"cached"] forState:UIControlStateNormal];
					});
				}];
			}
			else
			{
				NSUserDefaults *defaults = [NSUserDefaults standardUserDefaults];
				if (![defaults boolForKey:@"acknowledged-no-cache-on-3g"])
				{
					RIButtonItem *cancelButton = [RIButtonItem item];
					cancelButton.label = NSLocalizedString(@"Remind Me", @"");

					RIButtonItem *dismissButton = [RIButtonItem item];
					dismissButton.label = NSLocalizedString(@"Dismiss", @"");
					dismissButton.action = ^
					{
						[defaults setBool:YES forKey:@"acknowledged-no-cache-on-3g"];
					};

					UIAlertView *alertView = [[UIAlertView alloc] initWithTitle:NSLocalizedString(@"Caching Disabled", @"")
																		message:NSLocalizedString(@"Caching songs can only be done on WiFi networks.", @"")
															   cancelButtonItem:cancelButton otherButtonItems:dismissButton, nil];
					[alertView show];
					[alertView release];
				}
			}
        }
    }
}

- (void)toggleCacheForAlbum:(id)sender
{
    
}

- (void)updateNowPlayingImage:(id)sender
{
	[self.tableView reloadData];
}

- (void)setupHeaderAndFooter
{	
	UIColor* grayColor = [UIColor colorWithRed:0.9529f green:0.9529f blue:0.9529f alpha:1.0f];
	
	headerView = [[UIView alloc] initWithFrame:CGRectMake(0, 0, self.view.frame.size.width, 120)];
	headerView.backgroundColor = grayColor;
	
	coverArt = [[UIImageView alloc] initWithFrame:CGRectMake(10, 10, 100, 100)];
    NSError *error = nil;
    coverArt.image = [loader imageForArtID:self.album.coverArtId error:&error];
    
    if ([error code]) {
        // Let's see if we need to wait for non-dragging/decelerating
//        if (self.tableView.dragging == NO && self.tableView.decelerating == NO)
//        {
            [loader download:self.album.coverArtId];
//        }
    }

	[headerView addSubview:coverArt];
	
	artistLabel = [[UILabel alloc] initWithFrame:CGRectMake(120, 10, 200, 20)];
	artistLabel.font = [UIFont boldSystemFontOfSize:[UIFont systemFontSize]];
	artistLabel.backgroundColor = grayColor;
	artistLabel.shadowColor = [UIColor whiteColor];
	artistLabel.shadowOffset = CGSizeMake(0, 1);
	[headerView addSubview:artistLabel];
	
	artistLabel.text = self.artist;
	
	albumLabel = [[UILabel alloc] initWithFrame:CGRectMake(120, 30, 200, 20)];
	albumLabel.font = [UIFont boldSystemFontOfSize:[UIFont labelFontSize]];
	albumLabel.backgroundColor = grayColor;
	albumLabel.shadowColor = [UIColor whiteColor];
	albumLabel.shadowOffset = CGSizeMake(0, 1);
	[headerView addSubview:albumLabel];
	
	albumLabel.text = self.album.title;
	
    /* Not yet
    cacheButton = [UIButton buttonWithType:UIButtonTypeCustom];
    cacheButton.frame = CGRectMake(268, 76, 44, 44);
    // If none of the songs are cached, it should use uncached.png
    // If some of the songs are cached, it should use partiallycached.png
    // If all of the songs are cachced, it should use cached.png
    [cacheButton setImage:[UIImage imageNamed:@"uncached"] forState:UIControlStateNormal];
    [headerView addSubview:cacheButton];
     */
    
	self.tableView.tableHeaderView = headerView;
	
	self.tableView.backgroundColor = [UIColor colorWithRed:0.9529f green:0.9529f blue:0.9529f alpha:1.0f];
	
	self.tableView.tableFooterView = [[[UIView alloc] initWithFrame:CGRectZero] autorelease];
	UIView *bigFooterView = [[UIView alloc] initWithFrame:CGRectMake(0, 0, 320, 1000)];
	bigFooterView.backgroundColor = [UIColor whiteColor];
	bigFooterView.opaque = YES;
	[self.tableView.tableFooterView addSubview:bigFooterView];
	[bigFooterView release];
}

@end
