//
//  ArtistViewController.m
//  iSub
//
//  Created by Ben Baron on 2/28/10.
//  Copyright 2010 Ben Baron. All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without modification, 
//  are permitted provided that the following conditions are met:
//  
//  * Redistributions of source code must retain the above copyright notice, this 
//    list of conditions and the following disclaimer.
//  * Redistributions in binary form must reproduce the above copyright notice, 
//    this list of conditions and the following disclaimer in the documentation 
//    and/or other materials provided with the distribution.
//  * Neither the my name nor the names of my contributors may be used to endorse
//    or promote products derived from this software without specific prior written 
//    permission.
//  
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
//  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
//  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
//  SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
//  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED 
//  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR 
//  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
//  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
//  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH 
//  DAMAGE.

#import "ArtistViewController.h"
#import "AlbumViewController.h"
#import "AlbumUITableViewCell.h"
#import "Artist.h"
#import "Album.h"
#import "Song.h"


@interface ArtistViewController ()
- (void)configureCell:(UITableViewCell*)cell forAlbumAtIndexPath:(NSIndexPath*)indexPath;
@end


@implementation ArtistViewController
@synthesize artist, albums, songs;

- (id)initWithStyle:(UITableViewStyle)style
{
	if ((self = [super initWithStyle:style]))
	{
		self.title = NSLocalizedString(@"Albums",@"");
		self.albums = [NSMutableArray array];
		self.songs = [NSMutableArray array];
	}
	return self;
}

- (void)viewDidLoad
{
    [super viewDidLoad];
	self.tableView.rowHeight = 60.0f;
	self.title = self.artist.name;

	loader = [[AlbumArtLoader alloc] init];
	loader.delegate = self;
	loader.imageSize = 120;

	[self loadData:NO];
}

- (void)viewDidUnload
{
    [super viewDidUnload];
    loader.delegate = nil;
    [loader release], loader = nil;
}

#pragma mark -
#pragma mark Data Loader

- (void)loadLocalData
{
	@synchronized(self.albums)
	{
		[self.albums removeAllObjects];
		[PerThreadManagedObjectContext() refreshObject:self.artist mergeChanges:YES]; //force the album to reload its set of songs from the persistent store.
        NSSortDescriptor *descriptor = [NSSortDescriptor sortDescriptorWithKey:@"dearticlizedTitle" ascending:YES];
        NSArray *descriptors = [NSArray arrayWithObjects:descriptor, nil];
		[self.albums addObjectsFromArray:[self.artist.albums sortedArrayUsingDescriptors:descriptors]];
	}
	
//	@synchronized(self.songs)
//	{
//		[self.songs removeAllObjects];
//        NSSortDescriptor *descriptor = [NSSortDescriptor sortDescriptorWithKey:@"dearticlizedTitle" ascending:YES];
//        NSArray *descriptors = [NSArray arrayWithObjects:descriptor, nil];
//		[self.songs addObjectsFromArray:[self.artist.songs sortedArrayUsingDescriptors:descriptors]];
//	}
	
	[self.tableView reloadData];
}

- (void)beginLoadingRemoteData:(NSString*)force
{
	NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
	
	NSError *error = [self.artist loadAlbums];
	
	if (error)
	{
		[self performSelectorOnMainThread:@selector(failedLoadingData:) withObject:error waitUntilDone:NO];
	}
	
	[pool release];
	
	[self performSelectorOnMainThread:@selector(loadLocalData) withObject:nil waitUntilDone:YES];
	[self performSelectorOnMainThread:@selector(finishLoadingData) withObject:nil waitUntilDone:NO];
}

- (void)finishLoadingData
{
	[self.tableView reloadData];
	[super finishLoadingData];
}

#pragma mark -
#pragma mark Notifications

- (void)resetCredentials:(id)sender
{
	[super resetCredentials:sender];
	self.albums = [NSMutableArray array];
	self.songs = [NSMutableArray array];
	[self.tableView reloadData];
}

- (void)resetTableData:(id)sender
{
    [super resetTableData:sender];
    self.artist = nil;
    self.albums = [NSMutableArray array];
    self.songs = [NSMutableArray array];
    [self.tableView reloadData];
}

- (void)updateReachability:(Reachability *)reachability;
{
	[super updateReachability:reachability];
	
	for (AlbumUITableViewCell *cell in [self.tableView visibleCells])
	{
		[self configureCell:cell forAlbumAtIndexPath:[self.tableView indexPathForCell:cell]];
	}
}

#pragma mark Table view methods

- (NSInteger)numberOfSectionsInTableView:(UITableView *)tableView
{
	if ([self.songs count] > 0)
	{
		return 2;
	}
	else
	{
		return 1;
	}
}

- (NSInteger)tableView:(UITableView *)tableView numberOfRowsInSection:(NSInteger)section 
{
	if (0 == section)
	{
		return [self.albums count];
	}
	else
	{
		return [self.songs count];
	}
}

- (UITableViewCell *)tableView:(UITableView *)tableView cellForRowAtIndexPath:(NSIndexPath *)indexPath 
{	
	static NSString *AlbumCellIdentifier = @"AlbumCellIdentifier";
	AlbumUITableViewCell *cell = (AlbumUITableViewCell *)[tableView dequeueReusableCellWithIdentifier:AlbumCellIdentifier];
	
	if (nil == cell)
	{
		cell = [[[AlbumUITableViewCell alloc] initWithStyle:UITableViewCellStyleDefault reuseIdentifier:AlbumCellIdentifier] autorelease];
	}
	[self configureCell:cell forAlbumAtIndexPath:indexPath];
	return cell;
}

- (void)configureCell:(AlbumUITableViewCell*)cell forAlbumAtIndexPath:(NSIndexPath*)indexPath;
{
	cell.accessoryType = UITableViewCellAccessoryDisclosureIndicator;

	if (0 == indexPath.section)
	{
		Album *album = [self.albums objectAtIndex:indexPath.row];
		cell.albumNameLabel.text = album.title;

        NSError *error = nil;
        cell.coverArtView.image = [loader imageForArtID:album.coverArtId error:&error];
        
        if ([error code]) {
            // Let's see if we need to wait for non-dragging/decelerating
    //        if (self.tableView.dragging == NO && self.tableView.decelerating == NO)
    //        {
                [loader download:album.coverArtId];
    //        }
        }
		if (canStream || [album hasCachedSongs])
		{
			cell.albumNameLabel.textColor = [UIColor blackColor];
		}
		else
		{
			cell.albumNameLabel.textColor = [UIColor lightGrayColor];
			cell.accessoryType = UITableViewCellAccessoryNone;
		}
	}
	else
	{
		Song *song = [self.songs objectAtIndex:indexPath.row];
		cell.textLabel.text = song.title;
	}
}

- (void)tableView:(UITableView *)tableView didSelectRowAtIndexPath:(NSIndexPath *)indexPath
{
	if (0 == indexPath.section)
	{
		Album *selection = [self.albums objectAtIndex:indexPath.row];

		if (canStream || [selection hasCachedSongs])
		{
			AlbumViewController *albumViewController = [[[AlbumViewController alloc] initWithStyle:UITableViewStylePlain] autorelease];
			albumViewController.album = selection;
			albumViewController.artist = self.artist.name;
			[self.navigationController pushViewController:albumViewController animated:YES];
		}
		else
		{
			[tableView deselectRowAtIndexPath:indexPath animated:YES];
		}
	}
	else
	{
		Song *song = [self.songs objectAtIndex:indexPath.row];
		if (canStream || [song cachedSongExists])
		{
			[self didSelectSong:song withPlaylist:self.songs];
		}
		else
		{
			[tableView deselectRowAtIndexPath:indexPath animated:YES];
		}
	}
}

#pragma mark -
#pragma mark Memory Management

- (void)didReceiveMemoryWarning
{
	[super didReceiveMemoryWarning];
	self.albums = NULL;
}

- (void)dealloc
{
	self.artist = NULL;
	self.albums = NULL;
	self.songs = NULL;
	
	[super dealloc];
}

#pragma mark -
#pragma mark Deferred Image Loading

- (void)reloadImages
{
	[self.tableView reloadData];
}

- (void)loadImagesForOnscreenRows
{
	for (NSIndexPath *indexPath in [self.tableView indexPathsForVisibleRows])
	{
		Album *album = [self.tableData objectAtIndex:indexPath.row];
        // This will prompt the album art to download, but doesn't try to load the image when it's done?
		[loader download:album.coverArtId];
	}
}

- (void)scrollViewDidEndDragging:(UIScrollView *)scrollView willDecelerate:(BOOL)decelerate
{	
	[super scrollViewDidEndDragging:scrollView willDecelerate:decelerate];
	
    if (!decelerate)
	{
        [self loadImagesForOnscreenRows];
    }
}

- (void)scrollViewDidEndDecelerating:(UIScrollView *)scrollView
{	
    [self loadImagesForOnscreenRows];
}

@end
