//
//  PlaylistViewController.m
//  iSub
//
//  Created by Zachery Bir on 3/9/11.
//  Copyright 2011 Canonical Ltd.
//	
//  This program is free software: you can redistribute it and/or modify it
//  under the terms of the GNU Affero General Public License version 3,
//  as published by the Free Software Foundation.
//	
//  This program is distributed in the hope that it will be useful, but
//  WITHOUT ANY WARRANTY; without even the implied warranties of
//  MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
//  PURPOSE.  See the GNU Affero General Public License for more details.
//
//  You should have received a copy of the GNU Affero General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.

#import "PlaylistViewController.h"
#import "PlaylistEditViewController.h"
#import "Playlist.h"
#import "PlaylistParser.h"
#import "Song.h"
#import "NSMutableArray+Extras.h"
#import "RIButtonItem.h"
#import "UIAlertView+Blocks.h"

@interface PlaylistViewController ()
- (void)deleteLocalPlaylist;
@end

@implementation PlaylistViewController

@synthesize playlist, playlistListViewController, songs;

- (id)initWithStyle:(UITableViewStyle)style playlist:(Playlist *)aPlaylist
{
	if ((self = [super initWithStyle:style]))
	{
		self.title = self.playlist.name;
		self.playlist = aPlaylist;
		self.songs = [NSMutableArray array];
	}
	return self;
}

- (void)viewDidLoad
{
    [super viewDidLoad];
	
	UIView *header = [[UIView alloc] initWithFrame:CGRectMake(0,0,320,40)];
	header.autoresizingMask = UIViewAutoresizingFlexibleWidth;
	
	UIButton *editButton = [UIButton buttonWithType:UIButtonTypeRoundedRect];
	editButton.frame = CGRectMake(5, 5, 100, 30);
	[editButton setTitle:NSLocalizedString(@"Edit", @"") forState:UIControlStateNormal];
	[editButton addTarget:self action:@selector(editPlaylist) forControlEvents:UIControlEventTouchUpInside];
	[header addSubview:editButton];
	
	UIButton *clearButton = [UIButton buttonWithType:UIButtonTypeRoundedRect];
	clearButton.frame = CGRectMake(110, 5, 100, 30);
	[clearButton setTitle:NSLocalizedString(@"Clear", @"") forState:UIControlStateNormal];
	[clearButton addTarget:self action:@selector(clearPlaylist) forControlEvents:UIControlEventTouchUpInside];
	[header addSubview:clearButton];
	
	UIButton *deleteButton = [UIButton buttonWithType:UIButtonTypeRoundedRect];
	deleteButton.frame = CGRectMake(215, 5, 100, 30);
	[deleteButton setTitle:NSLocalizedString(@"Delete", @"") forState:UIControlStateNormal];
	[deleteButton addTarget:self action:@selector(deletePlaylist) forControlEvents:UIControlEventTouchUpInside];
	[header addSubview:deleteButton];
	
	if ([playlist isEditable])
		self.tableView.tableHeaderView = header;
	
	[header release];
}

#pragma mark -
#pragma mark Data Loader

- (void)loadLocalData
{
	@synchronized(self.songs)
	{
		[self.songs removeAllObjects];
		if ([self.playlist isKindOfClass:[NSManagedObject class]])
			[PerThreadManagedObjectContext() refreshObject:self.playlist mergeChanges:YES]; //force the playlist to reload its set of songs from the persistent store.
		[self.songs addObjectsFromArray:self.playlist.songs];
        // this should sort based on something else (like the index on the playlist or something)
//		[self.songs sortUsingSelector:@selector(compare:)];
	}
	
	[self.tableView reloadData];
}


- (void)beginLoadingRemoteData:(NSString*)force
{
	NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
	
	if (self.playlist.playlistId) // Protect against CouchDB access being down, at least allow for local playlists.
	{
		NSError *error = [self.playlist loadSongs];
	
		if (error)
		{
			[self performSelectorOnMainThread:@selector(failedLoadingData:) withObject:error waitUntilDone:NO];
		}
	}
	
	[self performSelectorOnMainThread:@selector(loadLocalData) withObject:nil waitUntilDone:NO];
	[self performSelectorOnMainThread:@selector(finishLoadingData) withObject:nil waitUntilDone:NO];
	[pool release];
}

- (void)didReceiveMemoryWarning
{
    [super didReceiveMemoryWarning];
}

- (void)dealloc
{
	self.playlist = nil;
    self.songs = nil;
	[super dealloc];
}

#pragma mark -
#pragma mark UISearchDisplayController Delegate Methods

- (BOOL)searchDisplayController:(UISearchDisplayController *)controller shouldReloadTableForSearchString:(NSString *)searchString
{	
	[self.searchResults removeAllObjects];
	
	for (Song* s in self.songs)
	{
		NSRange titleResultsRange = [s.title rangeOfString:searchString options:NSCaseInsensitiveSearch];
		
		if (titleResultsRange.length > 0)
			[self.searchResults addObject:s];
	}
	
	return YES;
}

#pragma mark - UITableView

- (NSInteger)tableView:(UITableView *)tableView numberOfRowsInSection:(NSInteger)section
{
	if (self.searchDisplayController.searchResultsTableView == tableView)
	{
		return [self.searchResults count];
	}
	else 
	{
		return [self.songs count];
	}
}

- (UITableViewCell *)tableView:(UITableView *)tableView cellForRowAtIndexPath:(NSIndexPath *)indexPath 
{
    static NSString *CellIdentifier = @"Cell";
    
    UITableViewCell *cell = [tableView dequeueReusableCellWithIdentifier:CellIdentifier];
    if (nil == cell)
    {
        cell = [[[UITableViewCell alloc] initWithStyle:UITableViewCellStyleSubtitle
                                       reuseIdentifier:CellIdentifier] autorelease];
    }

    Song *s = nil;
    
    if (self.searchDisplayController.searchResultsTableView == tableView)
    {
        s = [self.searchResults objectAtIndex:indexPath.row];
        cell.accessoryType = UITableViewCellAccessoryDisclosureIndicator;
    }
    else
    {
        s = [self.songs objectAtIndex:indexPath.row]; 
        cell.accessoryType = UITableViewCellAccessoryNone;
    }

    cell.textLabel.text = s.title;
    cell.detailTextLabel.text = [NSString stringWithFormat:@"%@ - %@", s.album, s.artist];

    return cell;
}

- (void)tableView:(UITableView *)tableView didSelectRowAtIndexPath:(NSIndexPath *)indexPath
{
	Song *song = nil;
    NSArray *selectPlaylist = nil;
		
	if (self.searchDisplayController.searchResultsTableView == tableView)
	{
		song = [self.searchResults objectAtIndex:indexPath.row];
        selectPlaylist = self.searchResults;
	}
	else 
	{	
		song = [self.songs objectAtIndex:indexPath.row];
        selectPlaylist = self.songs;
	}
		
	[self didSelectSong:song withPlaylist:selectPlaylist];
}

#pragma mark - Edit, Clear, Delete

- (void)editPlaylist
{
    PlaylistEditViewController *playlistEditViewController = [PlaylistEditViewController playlistEditViewControllerWithPlaylist:self.playlist originator:self];
    [self.navigationController presentModalViewController:playlistEditViewController animated:YES];
}

- (void)clearPlaylist
{
	RIButtonItem *cancelButton = [RIButtonItem item];
	cancelButton.label = NSLocalizedString(@"Cancel", @"");
	
	RIButtonItem *clearButton = [RIButtonItem item];
	clearButton.label = NSLocalizedString(@"Clear", @"");
	clearButton.action = ^
	{
		self.playlist.playlistSongIndexes = nil;
		[self.songs removeAllObjects];
		SaveContext();
		[self.tableView reloadData];
	};

	UIAlertView *alertView = [[UIAlertView alloc] initWithTitle:NSLocalizedString(@"Clear Playlist", @"")
														message:NSLocalizedString(@"This will clear this playlist. It cannot be undone.", @"")
											   cancelButtonItem:cancelButton otherButtonItems:clearButton, nil];
	[alertView show];
	[alertView release];
}

- (void)deletePlaylist
{
	RIButtonItem *cancelButton = [RIButtonItem item];
	cancelButton.label = NSLocalizedString(@"Cancel", @"");
	
	RIButtonItem *deleteButton = [RIButtonItem item];
	deleteButton.label = NSLocalizedString(@"Delete", @"");
	deleteButton.action = ^
	{
		if (self.playlist.playlistId)
		{
			NSMutableArray *extraParameters = [NSMutableArray array];
			[extraParameters addKeyValueObjectFromArray:[NSArray arrayWithObjects:@"id", self.playlist.playlistId, nil]];
			NSURL *url = [[Subsonic sharedSubsonic] getBaseURL:@"deletePlaylist.view" parameters:extraParameters];
			NSURLRequest *request = [[[NSURLRequest alloc] initWithURL:url] autorelease];
			NSURLConnection *connection = [NSURLConnection connectionWithRequest:request delegate:self];
			[connection start];
		}
		else
		{
			[self deleteLocalPlaylist];
		}
	};

	UIAlertView *alertView = [[UIAlertView alloc] initWithTitle:NSLocalizedString(@"Delete Playlist", @"")
														message:NSLocalizedString(@"This will delete this playlist. It cannot be undone.", @"")
											   cancelButtonItem:cancelButton otherButtonItems:deleteButton, nil];
	[alertView show];
	[alertView release];
}

- (void)connection:(NSURLConnection *)connection didReceiveResponse:(NSURLResponse *)response
{
	[self deleteLocalPlaylist];
}

- (void)deleteLocalPlaylist
{
	[PerThreadManagedObjectContext() deleteObject:self.playlist];
	SaveContext();
	[self.playlistListViewController performSelector:@selector(reload:) withObject:nil afterDelay:0.25];
	[self.navigationController popViewControllerAnimated:YES];
}

- (void)resetTableData:(id)sender
{
    [super resetTableData:sender];
    self.playlist = nil;
    [self.tableView reloadData];
}
@end

